/****************************************************************************
**  CUBE        http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2023                                                     **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#include "config.h"
#include <algorithm>
#include <vector>
#include <utility>
#include <stdint.h>
#include <iostream>
#include <sstream>
#include <string>
#include <future>

#include <CubeCnode.h>
#include "CubeIoProxy.h"
#include "POPCalculation.h"
#include "POPServerRequest.h"
#include "PerformanceAnalysis.h"
#include "PerformanceTest.h"
#include "POPAuditAnalysis.h"
#include "POPHybridAuditAnalysisAdd.h"
#include "POPHybridAuditAnalysis.h"
#include "BSPOPHybridAuditAnalysis.h"


using json = nlohmann::json;

std::vector<unsigned char>
popcalculation::POPCalculation::calculate( const std::vector<unsigned char>& request,
                                           cube::CubeIoProxy*                cube )
{
    popcalculation::PerformanceAnalysis* performance_analysis = nullptr;

    std::vector<cube::Cnode*> cnodev_ = cube->getCnodes();
    std::string               stringFromClient { request.begin(), request.end() };

    json j           = json::parse( stringFromClient );
    auto pop_request = j.template get<popserver_request::POPServerRequest>();

    switch ( pop_request.pop_analysis )
    {
        case HYBRID_ADD:
            performance_analysis = new hybaddanalysis::POPHybridAuditPerformanceAnalysisAdd( cube );
            break;
        case HYBRID_MULT:
            performance_analysis = new hybanalysis::POPHybridAuditPerformanceAnalysis( cube );
            break;
        case HYBRID_BSC:
            performance_analysis = new bscanalysis::BSPOPHybridAuditPerformanceAnalysis( cube );
            break;
        case PURE_MPI:
        default:
            performance_analysis = new mpianalysis::POPAuditPerformanceAnalysis( cube );
            break;
    }

    std::list<popcalculation::PerformanceTest*> pop_metrics     = performance_analysis->getPOPTests();
    std::list<popcalculation::PerformanceTest*> io_metrics      = performance_analysis->getIOTests();
    std::list<popcalculation::PerformanceTest*> add_metrics     = performance_analysis->getAdditionalTests();
    std::list<popcalculation::PerformanceTest*> control_metrics = performance_analysis->getControlTests();

    cube::list_of_cnodes _cnodes;
    for ( size_t i = 0; i < pop_request.cnodes.size(); ++i )
    {
        _cnodes.push_back(   std::make_pair( cnodev_[ pop_request.cnodes[ i ] ], cube::CalculationFlavour( pop_request.state[ i ] ) ) );
    }

// ----   ASYNC CALCULATION
    std::list<popcalculation::PerformanceTest*> all_tests = performance_analysis->getAllTestsForCalculation();

    std::vector<std::future<void> > calculation_done;
    auto                            asynx_calculate_metric = [ ]( const cube::list_of_cnodes& c,  popcalculation::PerformanceTest* t )
                                                             {
                                                                 if ( t->isActive() )
                                                                 {
                                                                     t->apply( c );
                                                                 }
                                                             };

    std::for_each( all_tests.begin(), all_tests.end(), [ &_cnodes, &calculation_done, &asynx_calculate_metric ]( popcalculation::PerformanceTest* t )
    {
        calculation_done.push_back(  std::async( std::launch::async, asynx_calculate_metric, _cnodes, t )  );
    } );
    std::for_each( calculation_done.begin(), calculation_done.end(), [ ]( std::future<void>& _f )
    {
        _f.get();     // waiting for the result
    } );
// ---- DONE, NOW COLLECT RESULTS






    popserver_request::POPServerAnswer pop_answer;
    pop_answer.pop_analysis_help = performance_analysis->getAnchorHowToMeasure();

    auto calculate_pop_metric = [ &pop_answer ]( popcalculation::PerformanceTest* t )
                                {
                                    pop_answer.pop_metric_names.push_back( t->name() );
                                    pop_answer.pop_metrics_helps.push_back( t->getHelpUrl() );
                                    if ( t != nullptr &&  t->isActive() )
                                    {
                                        double v = t->value();
                                        pop_answer.pop_values.push_back( v );
                                    }
                                    else
                                    {
                                        pop_answer.pop_values.push_back( -1 );
                                    }
                                };

    std::for_each( pop_metrics.crbegin(), pop_metrics.crend(), calculate_pop_metric );
    auto calculate_io_metric = [ &pop_answer ]( popcalculation::PerformanceTest* t )
                               {
                                   pop_answer.io_metric_names.push_back( t->name() );
                                   pop_answer.io_metrics_helps.push_back( t->getHelpUrl() );
                                   if ( t->isActive() )
                                   {
                                       double v = t->value();
                                       pop_answer.io_values.push_back( v );
                                   }
                                   else
                                   {
                                       pop_answer.io_values.push_back( -1 );
                                   }
                               };

    std::for_each( io_metrics.crbegin(), io_metrics.crend(), calculate_io_metric );
    auto calculate_add_metric = [ &pop_answer ]( popcalculation::PerformanceTest* t )
                                {
                                    pop_answer.additional_metric_names.push_back( t->name() );
                                    pop_answer.additional_metrics_helps.push_back( t->getHelpUrl() );
                                    if ( t->isActive() )
                                    {
                                        double v = t->value();
                                        pop_answer.additional_values.push_back( v );
                                    }
                                    else
                                    {
                                        pop_answer.additional_values.push_back( -1 );
                                    }
                                };

    std::for_each( add_metrics.crbegin(), add_metrics.crend(), calculate_add_metric );
    auto calculate_control_metric = [ &pop_answer ]( popcalculation::PerformanceTest* t )
                                    {
                                        pop_answer.control_metric_names.push_back( t->name() );
                                        pop_answer.control_metrics_helps.push_back( t->getHelpUrl() );
                                        if ( t->isActive() )
                                        {
                                            double min_v = t->min_value();
                                            double v     = t->value();
                                            double max_v = t->max_value();
                                            pop_answer.control_values.push_back( min_v );
                                            pop_answer.control_values.push_back( v );
                                            pop_answer.control_values.push_back( max_v );
                                        }
                                        else
                                        {
                                            pop_answer.control_values.push_back( -1 );
                                            pop_answer.control_values.push_back( -1 );
                                            pop_answer.control_values.push_back( -1 );
                                        }
                                    };

    std::for_each( control_metrics.crbegin(), control_metrics.crend(), calculate_control_metric );

    json j3 = pop_answer;

    std::stringstream sstr;
    sstr << j3;
    std::string json_answer = sstr.str();

    std::vector<unsigned char> messageFromServer;
    messageFromServer.insert( messageFromServer.begin(), json_answer.begin(), json_answer.end() );


    delete performance_analysis;
    return messageFromServer;
}
